// QUESTION #1: I made EventDispatcher.hpp a header-only library.
//  * What are the pros and cons of this approach?
//  * What are general good practices and when should we choose header-only libraries and when shouldn't we?

#include <array>
#include <algorithm> // for std::find
#include "ButtonEventArgs.hpp"

#pragma once

// QUESTION #2: EventDispatcher is here a template class that allows the user to select the
// EventArgs and the maximum number of handlers supported by the dispatcher.
//  * What are alternative design choices to avoid duplicated code in the presence of more events?
// I was thinking about creating a base (optionally abstract) EventArgs class to use within
// EventDispatcher, which can be inherited in order to specialize behavior - but that would have
// two main drawbacks:
// 1. we would have a single instance of EventDispatcher to subscribe to, and that would cause the
// size of the underlying callback array to be bigger, with consequent increase of memory footprint
// and execution time
// 2. the user would have to perform some pattern matching on the received event args to verify that
// it is the expected type
//  * What's the drawback of this approach compared to having EventArgs and array size hard-coded?
// On godbolt, the two approaches seem to compile to the same exact code for this use case,
// but when instantiating more different EventDispatchers in the main() function the assembly
// lines increase by ~40 each time.
template<typename EventArgs, size_t MaxNumberOfHandlers>
class EventDispatcher
{
public:
    using EventCallback = void (*)(const EventArgs&);

    // QUESTION #3: Is this ctor needed at all?
    // What difference does writing this ctor make?
    constexpr EventDispatcher() : callbacks_{} {}

    // QUESTION #4: Does the `inline` keyword make any difference?
    // The compiler should add `inline` on its own to avoid ODR issues.
    // Is there any reason to use the keyword `inline` when the definition is in the header, like here?
    inline bool RegisterCallback(EventCallback callback)
    {
        if (callback == nullptr) return false;

        const auto& alreadyRegisteredCallback = findCallback(callback);
        if (alreadyRegisteredCallback != callbacks_.end()) return true;

        const auto& firstFreeCallback = findCallback(nullptr);
        if (firstFreeCallback == callbacks_.end()) return false;

        *firstFreeCallback = callback;
        return true;
    }

    inline void UnregisterCallback(EventCallback callback)
    {
        auto registeredCallback = findCallback(callback);
        if (registeredCallback != callbacks_.end())
        {
            *registeredCallback = nullptr;
        }
    }

    inline void Dispatch(const EventArgs& e) const
    {
        for (const auto& callback : callbacks_)
        {
            if (callback)
            {
                callback(e);
            }
        }
    }

private:
    std::array<EventCallback, MaxNumberOfHandlers> callbacks_;

    inline typename std::array<EventCallback, MaxNumberOfHandlers>::iterator findCallback(EventCallback callback)
    {
        return std::find(callbacks_.begin(), callbacks_.end(), callback);
    }
};
